
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2013 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef _HDR_gsiClass
#define _HDR_gsiClass

#include "tlTypeTraits.h"

//  For a comprehensive documentation see gsi.h

namespace gsi
{

template <class X> 
void *_take_and_create (SerialArgs &args, tl::false_tag /*has_copy_ctor*/)
{
  tl_assert (false); // cannot copy object of this type
  return 0;
}

template <class X> 
void *_take_and_create (SerialArgs &from, tl::true_tag /*has_copy_ctor*/)
{
  return from.template take_object<X> ();
}

template <class X> 
void *_get_vector_of (SerialArgs &from, const ArgType &a, void *data, void (*cb) (void *data, void *obj), tl::false_tag /*has_copy_ctor*/) 
{
  tl_assert (false); // cannot copy object of this type
  return 0;
}

template <class X> 
void _get_vector_of (SerialArgs &from, const ArgType &a, void *data, void (*cb) (void *data, void *obj), tl::true_tag /*has_copy_ctor*/) 
{
  std::vector<X> vv;
  const std::vector<X> *v = &vv;
  if (a.is_cref) {
    v = &from.template get_value<const std::vector<X> &> (typename gsi::type_traits<const std::vector<X> &>::tag ());
  } else if (a.is_cptr) {
    v = from.template get_value<const std::vector<X> *> (typename gsi::type_traits<const std::vector<X> *>::tag ());
  } else if (a.is_ref) {
    v = &from.template get_value<std::vector<X> &> (typename gsi::type_traits<std::vector<X> &>::tag ());
  } else if (a.is_ptr) {
    v = from.template get_value<std::vector<X> *> (typename gsi::type_traits<std::vector<X> *>::tag ());
  } else {
    vv = from.template get_value< std::vector<X> > (typename gsi::type_traits< std::vector<X> >::tag ());
  }
  for (typename std::vector<X>::const_iterator o = v->begin (); o != v->end (); ++o) {
    (*cb) (data, new X (*o));
  }
}

template <class X> 
void _get_cptr_vector_of (SerialArgs &from, const ArgType &a, void *data, void (*cb) (void *data, void *obj))
{
  std::vector<const X *> vv;
  const std::vector<const X *> *v = &vv;
  if (a.is_cref) {
    v = &from.template get_value<const std::vector<const X *> &> (typename gsi::type_traits<const std::vector<const X *> &>::tag ());
  } else if (a.is_cptr) {
    v = from.template get_value<const std::vector<const X *> *> (typename gsi::type_traits<const std::vector<const X *> *>::tag ());
  } else if (a.is_ref) {
    v = &from.template get_value<std::vector<const X *> &> (typename gsi::type_traits<std::vector<const X *> &>::tag ());
  } else if (a.is_ptr) {
    v = from.template get_value<std::vector<const X *> *> (typename gsi::type_traits<std::vector<const X *> *>::tag ());
  } else {
    vv = from.template get_value< std::vector<const X *> > (typename gsi::type_traits< std::vector<const X *> >::tag ());
  }
  for (typename std::vector<const X *>::const_iterator o = v->begin (); o != v->end (); ++o) {
    (*cb) (data, (void *) *o);
  }
}

template <class X> 
void _get_ptr_vector_of (SerialArgs &from, const ArgType &a, void *data, void (*cb) (void *data, void *obj))
{
  std::vector<X *> vv;
  const std::vector<X *> *v = &vv;
  if (a.is_cref) {
    v = &from.template get_value<const std::vector<X *> &> (typename gsi::type_traits<const std::vector<X *> &>::tag ());
  } else if (a.is_cptr) {
    v = from.template get_value<const std::vector<X *> *> (typename gsi::type_traits<const std::vector<X *> *>::tag ());
  } else if (a.is_ref) {
    v = &from.template get_value<std::vector<X *> &> (typename gsi::type_traits<std::vector<X *> &>::tag ());
  } else if (a.is_ptr) {
    v = from.template get_value<std::vector<X *> *> (typename gsi::type_traits<std::vector<X *> *>::tag ());
  } else {
    vv = from.template get_value< std::vector<X *> > (typename gsi::type_traits< std::vector<X *> >::tag ());
  }
  for (typename std::vector<X *>::const_iterator o = v->begin (); o != v->end (); ++o) {
    (*cb) (data, *o);
  }
}

template <class X> 
void _destroy (X *x, tl::false_tag /*has public dtor*/)
{
  tl_assert (false); // cannot delete object of this type
}

template <class X> 
void _destroy (X *x, tl::true_tag /*has public dtor*/)
{
  delete x;
}

template <class X> 
void _push_vector_of (SerialArgs &to, const ArgType &a, std::list<TempObject> &temp_objects, const std::vector<void *> &objects, tl::false_tag /*has_copy_ctor*/) 
{
  tl_assert (false); // cannot copy object of this type
}

template <class X> 
void _push_vector_of (SerialArgs &to, const ArgType &a, std::list<TempObject> &temp_objects, const std::vector<void *> &objects, tl::true_tag /*has_copy_ctor*/) 
{
  tl_assert (a.inner != 0);

  std::vector<X> vv;
  std::vector<X> *v;
  if (a.is_ref || a.is_cref || a.is_ptr || a.is_cptr) {
    v = new std::vector<X> ();
    temp_objects.push_back (TempObject ());
    temp_objects.back ().set (v);
  } else {
    v = &vv;
  }

  v->reserve (objects.size ());
  for (std::vector<void *>::const_iterator o = objects.begin (); o != objects.end (); ++o) {
    v->push_back (*(X *)*o);
  }

  if (a.is_cref) {
    to.set_value (npod_cref_tag (), *v);
  } else if (a.is_cptr) {
    to.set_value (npod_cptr_tag (), v);
  } else if (a.is_ref) {
    to.set_value (npod_ref_tag (), *v);
  } else if (a.is_ptr) {
    to.set_value (npod_ptr_tag (), v);
  } else {
    to.set_value (npod_direct_tag (), vv);
  }
}

template <class X> 
void _push_cptr_vector_of (SerialArgs &to, const ArgType &a, std::list<TempObject> &temp_objects, const std::vector<void *> &objects)
{
  tl_assert (a.inner != 0);

  std::vector<const X *> vv;
  std::vector<const X *> *v;
  if (a.is_ref || a.is_cref || a.is_ptr || a.is_cptr) {
    v = new std::vector<const X *> ();
    temp_objects.push_back (TempObject ());
    temp_objects.back ().set (v);
  } else {
    v = &vv;
  }

  v->reserve (objects.size ());
  for (std::vector<void *>::const_iterator o = objects.begin (); o != objects.end (); ++o) {
    v->push_back ((const X *)*o);
  }

  if (a.is_cref) {
    to.set_value (npod_cref_tag (), *v);
  } else if (a.is_cptr) {
    to.set_value (npod_cptr_tag (), v);
  } else if (a.is_ref) {
    to.set_value (npod_ref_tag (), *v);
  } else if (a.is_ptr) {
    to.set_value (npod_ptr_tag (), v);
  } else {
    to.set_value (npod_direct_tag (), vv);
  }
}

template <class X> 
void _push_ptr_vector_of (SerialArgs &to, const ArgType &a, std::list<TempObject> &temp_objects, const std::vector<void *> &objects)
{
  tl_assert (a.inner != 0);

  std::vector<X *> vv;
  std::vector<X *> *v;
  if (a.is_ref || a.is_cref || a.is_ptr || a.is_cptr) {
    v = new std::vector<X *> ();
    temp_objects.push_back (TempObject ());
    temp_objects.back ().set (v);
  } else {
    v = &vv;
  }

  v->reserve (objects.size ());
  for (std::vector<void *>::const_iterator o = objects.begin (); o != objects.end (); ++o) {
    v->push_back ((X *)*o);
  }

  if (a.is_cref) {
    to.set_value (npod_cref_tag (), *v);
  } else if (a.is_cptr) {
    to.set_value (npod_cptr_tag (), v);
  } else if (a.is_ref) {
    to.set_value (npod_ref_tag (), *v);
  } else if (a.is_ptr) {
    to.set_value (npod_ptr_tag (), v);
  } else {
    to.set_value (npod_direct_tag (), vv);
  }
}

template <class X> 
void *_create (tl::false_tag)
{
  throw tl::Exception (tl::translate ("Object cannot be created here"));
  return 0;
}

template <class X> 
void *_create (tl::true_tag)
{
  return new X ();
}

template <class X> 
void _assign (tl::false_tag /*has_copy_ctor*/, void *, void *)
{
  throw tl::Exception (tl::translate ("Object cannot be copied here"));
}

template <class X> 
void _assign (tl::true_tag /*has_copy_ctor*/, void *dest, void *src)
{
  *(X *)dest = *(const X *)src;
}

template <class X> 
void _copy_to (SerialArgs &at, void *from, tl::false_tag /*has_copy_ctor*/) 
{
  throw tl::Exception (tl::translate ("Object cannot be copied here"));
}

template <class X> 
void _copy_to (SerialArgs &at, void *from, tl::true_tag /*has_copy_ctor*/) 
{
  at.set_value (typename type_traits<X>::tag (), *((const X *)from));
}

inline void *get_client_data (void *) 
{
  return 0;
}

inline void *get_client_data (gsi::ObjectBase *x) 
{
  return x->data ();
}

/**
 *  @brief The declaration of a specific class
 *
 *  This class declares all methods that are required to instantiate or copy an object
 *  or to call it's methods in some generic way.
 */
template <class X>
class Class
  : public ClassBase
{
public:
  typedef typename tl::type_traits<X>::has_copy_constructor has_copy_ctor;
  typedef typename tl::type_traits<X>::has_default_constructor has_default_ctor;
  typedef typename tl::type_traits<X>::has_public_destructor has_public_dtor;

  Class (const std::string &name, const Methods &mm, const std::string &doc = std::string ())
    : ClassBase (0, name, 0, doc, mm)
  {
    // ...
  }

  Class (const ClassBase &base, const std::string &name, const Methods &mm, const std::string &doc = std::string ())
    : ClassBase (0, name, &base, doc, mm)
  {
    // ...
  }

  Class (const std::string &name, const std::string &doc = std::string ())
    : ClassBase (0, name, 0, doc, Methods ())
  {
    // ...
  }

  Class (const ClassBase &base, const std::string &name, const std::string &doc = std::string ())
    : ClassBase (0, name, &base, doc, Methods ())
  {
    // ...
  }

  Class (const ObjectConnector &connector, const std::string &name, const Methods &mm, const std::string &doc = std::string ())
    : ClassBase (&connector, name, 0, doc, mm)
  {
    // ...
  }

  Class (const ObjectConnector &connector, const ClassBase &base, const std::string &name, const Methods &mm, const std::string &doc = std::string ())
    : ClassBase (&connector, name, &base, doc, mm)
  {
    // ...
  }

  Class (const ObjectConnector &connector, const std::string &name, const std::string &doc = std::string ())
    : ClassBase (&connector, name, 0, doc, Methods ())
  {
    // ...
  }

  Class (const ObjectConnector &connector, const ClassBase &base, const std::string &name, const std::string &doc = std::string ())
    : ClassBase (&connector, name, &base, doc, Methods ())
  {
    // ...
  }

  void *client_data (void *p) const
  {
    void *cd = basic_client_data (p);
    if (cd) {
      return cd;
    } else {
      X *x = (X *)p;
      return get_client_data (x);
    }
  }

  virtual void destroy (void *p) const
  {
    X *x = (X *)p;
    has_public_dtor hpd;
    _destroy (x, hpd);
  }

  virtual void *create () const
  {
    has_default_ctor cst;
    void *r = _create<X> (cst);
    return r;
  }

  virtual void assign (void *dest, void *src) const
  {
    has_copy_ctor cst;
    _assign<X> (cst, dest, src);
  }

  virtual void get_vector_of (SerialArgs &from, const ArgType &a, void *data, void (*cb) (void *data, void *obj)) const 
  {
    has_copy_ctor cpt;
    tl_assert (a.inner != 0);
    if (a.inner->is_cptr) {
      _get_cptr_vector_of<X> (from, a, data, cb);
    } else if (a.inner->is_ptr) {
      _get_ptr_vector_of<X> (from, a, data, cb);
    } else {
      _get_vector_of<X> (from, a, data, cb, cpt);
    }
  }

  void push_vector_of (SerialArgs &to, const ArgType &a, std::list<TempObject> &temp_objects, const std::vector<void *> &objects) const 
  {
    has_copy_ctor cpt;
    tl_assert (a.inner != 0);
    if (a.inner->is_cptr) {
      _push_cptr_vector_of<X> (to, a, temp_objects, objects);
    } else if (a.inner->is_ptr) {
      _push_ptr_vector_of<X> (to, a, temp_objects, objects);
    } else {
      _push_vector_of<X> (to, a, temp_objects, objects, cpt);
    }
  }

  virtual void *take_and_create (SerialArgs &from) const
  {
    has_copy_ctor cpt;
    return _take_and_create<X> (from, cpt);
  }

  virtual void copy_to (SerialArgs &at, void *from) const 
  {
    has_copy_ctor cpt;
    _copy_to<X> (at, from, cpt);
  }

  virtual void *create_obj_iterator (SerialArgs &from) const
  {
    return from.template get_value<gsi::IterAdaptorAbstractBase *> (typename type_traits <gsi::IterAdaptorAbstractBase *>::tag ());
  }

  virtual bool obj_iterator_at_end (void *iter) const
  {
    gsi::IterAdaptorAbstractBase *i = (gsi::IterAdaptorAbstractBase *) iter;
    return i->at_end ();
  }

  virtual void *take_iterator_obj (void *iter) const
  {
    gsi::IterAdaptorAbstractBase *i = (gsi::IterAdaptorAbstractBase *) iter;
    return i->v_get_copy ();
  }

  virtual const void *get_iterator_obj (void *iter) const
  {
    gsi::IterAdaptorAbstractBase *i = (gsi::IterAdaptorAbstractBase *) iter;
    return i->v_get ();
  }

  virtual void inc_obj_iterator (void *iter) const
  {
    gsi::IterAdaptorAbstractBase *i = (gsi::IterAdaptorAbstractBase *) iter;
    i->inc ();
  }

  virtual void delete_obj_iterator (void *iter) const
  {
    gsi::IterAdaptorAbstractBase *i = (gsi::IterAdaptorAbstractBase *) iter;
    delete i;
  }

  virtual ClassBase *declaration () const
  {
    return &decl;
  }

  virtual bool can_destroy () const
  {
    has_public_dtor hpd;
    return tl::value_of (hpd);
  }

  virtual bool can_copy () const
  {
    has_copy_ctor cpt;
    return tl::value_of (cpt);
  }

  virtual bool can_default_create () const
  {
    has_default_ctor cpt;
    return tl::value_of (cpt);
  }

  virtual const ClassBase *subclass_decl (const void *p) const 
  {
    try {
      const X *x = (const X *)p;
      const std::type_info &ti_base = typeid (*x);
      const ClassBase *cls = find_subclass_decl (ti_base);
      return cls ? cls : this;
    } catch (std::bad_typeid &) {
      return this;
    }
  }

  virtual const ClassBase *find_subclass_decl (const std::type_info &ti) const 
  {
    if (ti == typeid (X)) {
      return this;
    }

    for (std::vector <const ClassBase *>::const_iterator s = subclasses ().begin (); s != subclasses ().end (); ++s) {
      const ClassBase *cls = (*s)->find_subclass_decl (ti);
      if (cls) {
        return cls;
      }
    }

    return 0;
  }

  static Class<X> decl;
};

}

#endif


