
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2013 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef HDR_gsiExpression
#define HDR_gsiExpression

#include "tlExpression.h"

#include "config.h"

#include <string>
#include <map>

namespace gsi 
{

class ClassBase;
template <class T> class Class;

/**
 *  @brief A EvalClass and VariantUserClassBase specialization that links GSI classes, Eval classes and Variant classes
 */
class KLAYOUT_DLL VariantUserClassBase
  : public tl::VariantUserClassBase, public tl::EvalClass
{
public:
  VariantUserClassBase (const gsi::ClassBase *cls, const VariantUserClassBase *object_cls, bool reference, bool is_const);
  ~VariantUserClassBase ();

  virtual void *create () const;
  virtual void destroy (void *obj) const;
  virtual bool equal (void *, void *) const;
  virtual bool less (void *, void *) const;
  virtual void *clone (void *obj) const;
  virtual std::string to_string (void *) const;
  virtual unsigned int type_code () const;
  virtual const char *name () const;
  virtual void read (void *, tl::Extractor &ex) const;
  virtual void execute (const tl::ExpressionParserContext &context, tl::Variant &out, tl::Variant &object, const std::string &method, const std::vector<tl::Variant> &args) const;

  const gsi::ClassBase *gsi_cls () const
  {
    return mp_cls;
  }

  bool is_class () const 
  { 
    return mp_object_cls != 0; 
  }

  bool is_ref () const 
  { 
    return m_reference; 
  }

  bool is_const () const 
  { 
    return m_is_const; 
  }

private:
  const gsi::ClassBase *mp_cls;
  const VariantUserClassBase *mp_object_cls;
  bool m_reference, m_is_const;
  std::map <std::string, int> m_method_ids;

  virtual void execute_gsi (const tl::ExpressionParserContext &context, tl::Variant &out, tl::Variant &object, const std::string &method, const std::vector<tl::Variant> &args) const;
};

/**
 *  @brief Find the VariantUserClassBase for a given GSI class
 */
KLAYOUT_DLL const VariantUserClassBase *
find_class (const gsi::ClassBase *gsi_cls, bool ref, bool is_const);

/**
 *  @brief Utility: initialize a variant from GSI type reference
 */
template <class T>
void make_variant_ref (tl::Variant &v, T *t)
{
  const VariantUserClassBase *c = find_class (&gsi::Class<T>::decl, true, false);
  tl_assert (c != 0);
  v.set_user (0, t, c);
}

/**
 *  @brief Utility: initialize a variant from GSI type reference
 */
template <class T>
void make_variant_ref (tl::Variant &v, const T *t)
{
  const VariantUserClassBase *c = find_class (&gsi::Class<T>::decl, true, true);
  tl_assert (c != 0);
  v.set_user (0, t, c);
}

/**
 *  @brief Utility: initialize a variant from GSI type (take over ownership)
 */
template <class T>
void make_variant (tl::Variant &v, T *t)
{
  const VariantUserClassBase *c = find_class (&gsi::Class<T>::decl, false, false);
  tl_assert (c != 0);
  v.set_user (0, t, c);
}

/**
 *  @brief Utility: initialize a variant from GSI type (deep copy)
 */
template <class T>
void make_variant (tl::Variant &v, const T &t)
{
  const VariantUserClassBase *c = find_class (&gsi::Class<T>::decl, false, false);
  tl_assert (c != 0);
  v.set_user (0, new T(t), c);
}

/**
 *  @brief Initialize GSI objects for expressions
 *
 *  This function must be called initially to enable GSI objects into expressions.
 */
KLAYOUT_DLL void
initialize_expressions ();

}

#endif

