
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2013 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "BrowserDialog.h"
#include "layBrowserPanel.h"
#include "layBrowserDialog.h"
#include "layMainWindow.h"
#include "layExceptions.h"

namespace lay 
{

BrowserDialog::BrowserDialog ()
  //  like all the other dialogs, we do not specify a parent, so the non-modal window will no stay in 
  //  front of the main window (but: then any QInputDialog opened in the loadUrl method will push
  //  the browser dialog to the back since it refers to the main window ..)
  : QDialog (0)
{
  Ui::BrowserDialog::setupUi (this);

  setObjectName (QString::fromAscii ("html_browser"));
  show ();
}

BrowserDialog::~BrowserDialog ()
{
  set_source (0);
}

void 
BrowserDialog::load (const std::string &s)
{
  browser->load (s);
}

void 
BrowserDialog::set_source (BrowserSource *source)
{
  browser->set_source (source);
}

void 
BrowserDialog::set_home (const std::string &url)
{
  browser->set_home (url);
}

void 
BrowserDialog::reload ()
{
  browser->reload ();
}

void
BrowserDialog::accept ()
{
  BEGIN_PROTECTED
  closed ();
  END_PROTECTED
  QDialog::accept ();
}

}

